/*
    Aquaponic IoT Controler
    Copyright (C) 2015  Davor Marjanović <aquaponic.controller@gmail.com>
    Licence: GNU GENERAL PUBLIC LICENSE
*/

/*jslint node:true, vars:true, bitwise:true, unparam:true */
/*jshint unused:true */

var mraa = require('mraa');
console.log('MRAA Version: ' + mraa.getVersion());
var express = require('express');
var app = express();
var path = require('path');
var http = require('http').Server(app);
var io = require('socket.io')(http);

var periodicInterval;
var logMsg = "";
// LEDs
// System on
var powerLed = new mraa.Gpio(12);
powerLed.dir(mraa.DIR_OUT);
// Led indicating something is wrong with the water flow
var warningLedFlow = new mraa.Gpio(10);
warningLedFlow.dir(mraa.DIR_OUT);
var warningLedFlowState = false;
// Led indicationg something is wrong with the Light system
var warningLedLight = new mraa.Gpio(11);
warningLedLight.dir(mraa.DIR_OUT);
var warningLedLightState = false;

var upmBuzzer = require("jsupm_buzzer");// Initialize on GPIO 5
var myBuzzer = new upmBuzzer.Buzzer(9);
myBuzzer.setVolume(0.5);
var toneCycleIndex = 0;
var buzzerInterval;
var buzzerActive = false;

// RELAY setup
// Relay 1 (Water Pump)
var relay1 = new mraa.Gpio(4);
relay1.dir(mraa.DIR_OUT);
relay1.write(1);
var relay1State = false;
var pauseRelay1 = false;
var pauseRelay1End;
// Relay 2 (Lights)
var relay2 = new mraa.Gpio(5);
relay2.dir(mraa.DIR_OUT);
relay2.write(1);
var relay2State = false;
var relay2LastChange = new Date();
// Relay 3 (Air pump)
var relay3 = new mraa.Gpio(6);
relay3.dir(mraa.DIR_OUT);
relay3.write(1);
var relay3State = false;
// Relay 4 (Fan)
var relay4 = new mraa.Gpio(7);
relay4.dir(mraa.DIR_OUT);
relay4.write(1);
var relay4State = false;

// WATER FLOW setup
var waterFlow_lib = require('jsupm_grovewfs');
// Instantiate a Grove Water Flow Sensor on digital pin D2 (note: only D2 and D3 pins can be used)
var flowSensor = new waterFlow_lib.GroveWFS(2);
var flowCalibrationFactor = 5.5;
var flowCountOld = 0;
var flowRate = 0.0;
var flowMilliLiters = 0;
var flowLiters = 0;
var flowOldTime = 0;
var flowTimeStart = new Date();
var flowTimeLast  = 0;
var flowTimeAvg   = 0;
var flowTimeCnt   = 0;
// set the flow counter to 0 and start counting
flowSensor.clearFlowCounter();
flowSensor.startFlowCounter();

// Temperature & Light Sensor Setup
var temperaturePin = new mraa.Aio(0); //Analog pin A0
var lightPin       = new mraa.Aio(1); //Analog pin A1
var lightPinSys    = new mraa.Aio(2); //Analog pin A2
var temperatureArr = [0,0,0,0,0,0,0,0,0,0]; // smoothing array
var lightArr       = [0,0,0,0,0,0,0,0,0,0]; // smoothing array
var lightSysArr    = [0,0,0,0,0,0,0,0,0,0]; // smoothing array
var temperature    = 20.0;
var light          = 50.0;
var lightSys       = 50.0;
var lightBuffer    = 1.0;

// SETTINGS
var litersPerCycle, pauseCycle, lightStartHour, lightEndHour, lightMinimum, airPumpConnected, fanConnected, temperatureMax;
var fs = require('fs');
loadSettings();

app.get('/', function(req, res) {res.sendFile(path.join(__dirname + '/webapp', 'index.html'));});
app.use(express.static(__dirname + '/webapp'));
app.use('/webapp', express.static(__dirname + '/webapp'));
//Socket.io Event handlers
io.on('connection', function(socket) {
    fetchSettings();
    sendData();
    
    socket.on('fetch settings', function(msg) {fetchSettings();});
    
    socket.on('store settings', function(msg) {
        storeSettings(msg);
        if (msg.litersPerCycle != litersPerCycle) resetFlow();
        loadSettings();
        setTimeout(fetchSettings, 1000);
    });
    
    socket.on('sound stop', function(msg){clearInterval(buzzerInterval); myBuzzer.stopSound();});
    socket.on('reset flow', function(msg){resetFlow();});
});
http.listen(3000, function(){console.log('Web server Active listening on *:3000');});


var checkDate = checkDate = new Date();
var initializationLoop = 0;
systemInitialization();

function systemInitialization(){
    if (initializationLoop == 0){ // first pause for 10 seconds for sensors to stabilize
        initializationLoop++;
        setTimeout(systemInitialization, 10000);
    }else if (initializationLoop <= 10){// collect starting ambient data
        checkAmbient();
        sendData();
        initializationLoop++;
        setTimeout(systemInitialization, 1000);
    }else{
        powerLed.write(1);
        // Initialize periodic functions
        periodicInterval = setInterval(function(){
            checkDate = new Date();
            measureFlow();
            checkAmbient();
            relayCheck();
            alertCheck();
            sendData();
            logMsg = "";
        }, 1000);
    }
}

function alertCheck(){
    if (relay1State && flowRate == 0){
        if (!warningLedFlowState){
            warningLedFlow.write(1);
            warningLedFlowState = true;
            addMsg("Flow Alert active.");
            if(!buzzerActive){
                buzzerInterval = setInterval(function(){alarm(1,10);}, 100);
                buzzerActive = true;
            }
        }
    } else if (warningLedFlowState) {
        warningLedFlow.write(0);
        warningLedFlowState = false;
        if (buzzerActive  && !warningLedLightState){
            clearInterval(buzzerInterval);
            myBuzzer.stopSound();
            buzzerActive = false;
            addMsg("Flow Alert passed.");
        }
    }
    
    if (relay2State && lightSys < lightMinimum){
        if (!warningLedLightState){
            warningLedLight.write(1);
            warningLedLightState = true;
            addMsg("Light Alert active.");
            if(!buzzerActive){
                buzzerInterval = setInterval(function(){alarm(1,10);}, 100);
                buzzerActive = true;
            }
        }
    } else if (warningLedLightState) {
        warningLedLight.write(0);
        warningLedLightState = false;
        if (buzzerActive  && !warningLedFlowState){
            clearInterval(buzzerInterval);
            myBuzzer.stopSound();
            buzzerActive = false;
            addMsg("Light Alert passed.");
        }
    }
}

function addMsg(msg){
    if (logMsg == ""){
        logMsg = msg;
    }else{
        logMsg += "<br>" + msg;
    }
}

function getMsg(){
    if (logMsg == ""){
        return "";
    }else{
        return "<span>"+checkDate.toLocaleTimeString()+": </span>"+logMsg;
    }
}

function sendData(){
    io.emit('send data', {light: light.toFixed(0), lightSys: lightSys.toFixed(0), temp: temperature.toFixed(0), flowRate: flowRate.toFixed(2), flowLiters: flowLiters.toFixed(2)
                        , relay1: relay1State, relay2: relay2State, relay3: relay3State, relay4: relay4State
                        , event: getMsg(), flowTimeLast: flowTimeLast, flowTimeAvg: flowTimeAvg, flowTimeCnt: flowTimeCnt});
}

// Check Relay Rules
function relayCheck(){   
    var current_hour = checkDate.getHours();
    var current_min  = checkDate.getMinutes();
    
    // Flow control
    if (flowLiters < litersPerCycle && !pauseRelay1){
        if (!relay1State){
            relay1.write(0);
            relay1State = true;
            flowTimeStart = checkDate;
            console.log('Water Pump ON');
            addMsg("Water Pump: ON");
        }
    }else{
        if(relay1State){
            relay1.write(1);
            relay1State = false;
            pauseRelay1 = true;
            
            flowTimeLast = parseInt(((checkDate.getTime() - flowTimeStart.getTime()) / 1000).toFixed(0));
            if (flowTimeCnt == 0){
                flowTimeAvg = parseInt(flowTimeLast);
            }else{
                flowTimeAvg = parseInt(((flowTimeAvg * flowTimeCnt) + flowTimeLast) / (flowTimeCnt + 1));
            }
            flowTimeCnt++;
            
            pauseRelay1End = new Date(checkDate.getTime() + pauseCycle*60000);
            console.log('Water Pump OFF Pause Initiated');
            addMsg("Water Pump: OFF, Reinitiate at: " + pauseRelay1End.toLocaleTimeString());
        } else if (pauseRelay1 && checkDate > pauseRelay1End){
            pauseRelay1 = false;
            flowMilliLiters = 0;
            flowLiters = 0;
        }
    }
    
    // Light control
    var lightPeriod = (lightEndHour > lightStartHour && current_hour >= lightStartHour && current_hour < lightEndHour) ||
                      (lightEndHour < lightStartHour && (current_hour >= lightStartHour || current_hour < lightEndHour));
    lightBuffer = arrayMax(lightArr) - arrayMin(lightArr);
    if (light < (lightMinimum - lightBuffer) && lightPeriod){
        if (!relay2State){
            relay2.write(0);
            relay2State = true;
            relay2LastChange = checkDate;
            console.log('Lights ON');
            addMsg("Lights: ON, Current Ambient Light: " + light.toFixed(0));
        }
    }else if (light > (lightMinimum + lightBuffer) || !lightPeriod){
        if(relay2State){
            relay2.write(1);
            relay2State = false;
            relay2LastChange = checkDate;
            console.log('Lights OFF');
            addMsg("Lights: OFF, Current Ambient Light: " + light.toFixed(0));
        }
    }
    
    // Air pump control
    if (airPumpConnected == 1){
        if (!relay3State){
            relay3.write(0);
            relay3State = true;
            console.log('Air Pump ON');
            addMsg("Air Pump turned ON.");
        }
    }else{
        if(relay3State){
            relay3.write(1);
            relay3State = false;
            console.log('Air Pump OFF');
            addMsg("Air Pump turned OFF.");
        }
    }
    
    // Fan control
    if (fanConnected == 1 && temperature > temperatureMax){
        if (!relay4State){
            relay4.write(0);
            relay4State = true;
            console.log('Fan ON');
            addMsg("Fan turned ON, current temperature is: "+ temperature.toFixed(0) +".");
        }
    }else{
        if(relay4State){
            relay4.write(1);
            relay4State = false;
            console.log('Fan OFF');
            addMsg("Fan turned OFF, current temperature is: "+ temperature.toFixed(0) +".");
        }
    }
}

// WATER FLOW
function measureFlow(){
	millis = flowSensor.getMillis();
	flowCount = flowSensor.flowCounter();

	fr = flowSensor.flowRate();
    flowRate = ((1000.0 / (millis - flowOldTime)) * (flowCount - flowCountOld)) / flowCalibrationFactor;
    flowOldTime = millis;
    var flowMilliLitersTemp = (flowRate / 60) * 1000;
    flowMilliLiters += flowMilliLitersTemp;
    var flowLitersTemp = flowMilliLitersTemp / 1000;
    flowLiters = flowMilliLiters / 1000;
    flowCountOld = flowCount;

	// output milliseconds passed, flow count, and computed flow rate
	/*outputStr = "Millis: " + millis + " Flow Count: " + flowCount +
	" Flow Rate: " + flowRate.toFixed(2) + " LPM, Interval Flow: " + flowLitersTemp.toFixed(2) + "L Total Flow: " + flowLiters.toFixed(2) +"L";
	console.log(outputStr);*/
}
function resetFlow(){
    flowTimeCnt  = 0;
    flowTimeLast = 0;
    flowTimeAvg  = 0;
}

// Ambient temperature and light
function averageArray(elmt) {
    var sum = 0;
    for( var i = 0; i < elmt.length; i++ ){
        sum += elmt[i];
    }
    var avg = sum/elmt.length;
    return avg;
}
function checkAmbient() {
    var temperatureValue = temperaturePin.read();
    var volts = (temperatureValue/1024.0)*5.0;
    var temp = (volts - .5) * 100;
    temperatureArr.shift();
    temperatureArr.push(temp);
    temperature = averageArray(temperatureArr);
    
    var lightValue       = lightPin.read();
    temp = (100/1024) * lightValue; // %
    lightArr.shift();
    lightArr.push(temp);
    light = averageArray(lightArr);
    
    var lightSysValue     = lightPinSys.read();
    temp = (100/1024) * lightSysValue; // %
    lightSysArr.shift();
    lightSysArr.push(temp);
    lightSys = averageArray(lightSysArr);
    
    //console.log("Temp:"+ temperature +"°C  Light:"+ light);
    //io.emit("temperature", temperature.toFixed(2));
    //io.emit("light", light.toFixed(2));
}

// Load & Store Settings to settings.txt
function loadSettings(){
    var content;
    fs.readFile('/node_app_slot/setting.txt', 'utf8', function read(err, data) {
        if (err) {
            throw err;
        }
        console.log(data);
        content = data.split(";");
        // litersPerCycle, pauseCycle, lightStartHour, lightEndHour, lightMinimum, airPumpConnected, fanConnected, temperatureMax
        litersPerCycle   = parseInt(content[0]);
        pauseCycle       = parseInt(content[1]);
        lightStartHour   = parseInt(content[2]);
        lightEndHour     = parseInt(content[3]);
        lightMinimum     = parseInt(content[4]);
        airPumpConnected = parseInt(content[5]);
        fanConnected     = parseInt(content[6]);
        temperatureMax   = parseInt(content[7]);
        flowTimeCnt      = 0;

        console.log("litersPerCycle:"       + content[0]);
        console.log("pauseCycle(minutes):"  + content[1]);
        console.log("lightStartHour:"       + content[2]);
        console.log("lightEndHour:"         + content[3]);
        console.log("lightMinimum:"         + content[4]);
        console.log("airPumpConnected:"     + content[5]);
        console.log("fanConnected:"         + content[6]);
        console.log("temperatureMax:"       + content[7]);
    });
}

function storeSettings(msg){
    var settingString = msg.litersPerCycle +";"+ 
                        msg.pauseCycle +";"+ 
                        msg.lightStartHour +";"+ 
                        msg.lightEndHour +";"+ 
                        msg.lightMinimum +";"+ 
                        msg.airPumpConnected +";"+ 
                        msg.fanConnected +";"+ 
                        msg.temperatureMax;
    fs.writeFile('/node_app_slot/setting.txt', settingString, function (err) {
      if (err) return console.log(err);
      console.log('settings.txt > '+ settingString);
    });
    //loadSettings(settingString);
}

function fetchSettings(){
    io.emit('fetch settings', {litersPerCycle: litersPerCycle,
                                   pauseCycle: pauseCycle,
                                   lightStartHour: lightStartHour,
                                   lightEndHour: lightEndHour,
                                   lightMinimum: lightMinimum,
                                   airPumpConnected: airPumpConnected,
                                   fanConnected: fanConnected,
                                   temperatureMax: temperatureMax});
}

//buzzer alarm
function alarm(toneCycles, totalCycles){
    if (toneCycleIndex < toneCycles){
        myBuzzer.playSound(3300, 100);
    }else{
        myBuzzer.stopSound();
    }
    toneCycleIndex++;
    //Reset the sound to start from the beginning. 
    if (toneCycleIndex == totalCycles) toneCycleIndex = 0;
}

// Util functions
function arrayMin(arr) {
  var len = arr.length, min = Infinity;
  while (len--) {
    if (arr[len] < min) {
      min = arr[len];
    }
  }
  return min;
};
function arrayMax(arr) {
  var len = arr.length, max = -Infinity;
  while (len--) {
    if (arr[len] > max) {
      max = arr[len];
    }
  }
  return max;
};
function getMinSec(sec){
    var min = (sec / 60).toFixed(0);
    return min + ":"+ (sec % 60);
}

// When exiting: clear interval and print message
process.on('SIGINT', function()
{
	clearInterval(periodicInterval);
	flowSensor.stopFlowCounter();
	flowSensor = null;
	waterFlow_lib.cleanUp();
	waterFlow_lib = null;
    myBuzzer.stopSound();
    powerLed.write(0);

	console.log("Exiting");
	process.exit(0);
});